#include <stdlib.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include "guiutils.h"
#include "guirgbimg.h"
#include "headerdlg.h"
#include "config.h"

#include "images/icon_info_20x20.xpm"
#include "images/icon_edit_32x32.xpm"
#include "images/icon_edit_48x48.xpm"
#include "images/icon_ok_20x20.xpm"
#include "images/icon_cancel_20x20.xpm"


typedef struct _header_dlg_struct	header_dlg_struct;


/*
 *	Header Dialog:
 */
struct _header_dlg_struct {
	GtkWidget	*toplevel;
	GtkAccelGroup	*accelgrp;
	gint		freeze_count;

	GtkWidget	*title_entry,
			*author_entry,
			*comments_entry,
			*x_spin,
			*y_spin,
			*base_width_spin,
			*base_height_spin,
			*preview_da;

	GdkPixmap	*preview_pm;

	gint		preview_button,
			preview_last_x,
			preview_last_y;

	gfloat		preview_dtow_coeff;

	/* Original image (shared) */
	const guint8	*rgba;
	gint		width, height, bpl;

	/* Thumb image for base size preview */
	guint8		*thumb_rgba;
	gint		thumb_width, thumb_height;

};
#define HEADER_DLG(p)	((header_dlg_struct *)(p))


static gint close_cb(GtkWidget *widget, GdkEvent *event, gpointer data);
static void set_btn_cb(GtkWidget *widget, gpointer data);
static void cancel_btn_cb(GtkWidget *widget, gpointer data);
static void adj_value_changed_cb(GtkAdjustment *adj, gpointer data);
static gint preview_event_cb(GtkWidget *widget, GdkEvent *event, gpointer data);
static void preview_resize(header_dlg_struct *d);
static void preview_draw(header_dlg_struct *d);

static void build_std_dialog(
	gint width, gint height, const gchar *title,
	GtkWidget *ref_toplevel,
	GtkWidget **toplevel_rtn, GtkWidget **client_vbox_rtn,
	GtkAccelGroup **accelgrp_rtn
);

gboolean iv_edit_header(
	GtkWidget *ref_toplevel,
	const gchar *filename,
	const guint8 *rgba,
	gint width, gint height, gint bpl,
	gchar **title, gchar **author, gchar **comments,
	gint *x, gint *y,
	gint *base_width, gint *base_height 
);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? g_strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : TRUE)


static gboolean		got_response;


/*
 *	Toplevel GtkWindow "delete_event" signal callback.
 */
static gint close_cb(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	cancel_btn_cb(widget, data);
	return(TRUE);
}

/*
 *	Set callback.
 */
static void set_btn_cb(GtkWidget *widget, gpointer data)
{
	GtkWidget *toplevel = (GtkWidget *)data;
	if(toplevel == NULL)
	    return;

	got_response = TRUE;

	gtk_main_quit();
}

/*
 *	Cancel callback.
 */
static void cancel_btn_cb(GtkWidget *widget, gpointer data)
{
	GtkWidget *toplevel = (GtkWidget *)data;
	if(toplevel == NULL)
	    return;

	got_response = FALSE;

	gtk_main_quit();
}

/*
 *	GtkAdjustment "value_changed" signal callback.
 */
static void adj_value_changed_cb(GtkAdjustment *adj, gpointer data)
{
	GtkWidget *w;
	header_dlg_struct *d = HEADER_DLG(data);
	if(d == NULL)
	    return;

	if(d->freeze_count > 0)
	    return;
	else
	    d->freeze_count++;

	w = d->preview_da;
	if(w != NULL)
	{
	    preview_resize(d);
	    gtk_widget_queue_draw(w);
	}

	d->freeze_count--;
}


/*
 *	Preview GtkDrawingArea event signal callback.
 */
static gint preview_event_cb(GtkWidget *widget, GdkEvent *event, gpointer data)
{
	gint status = FALSE;
	GdkEventConfigure *configure;
	GdkEventExpose *expose;
	GdkEventButton *button;
	GdkEventMotion *motion;
	header_dlg_struct *d = HEADER_DLG(data);
	if((widget == NULL) || (event == NULL) || (d == NULL))
	    return(status);

	switch((gint)event->type)
	{
	  case GDK_CONFIGURE:
	    configure = (GdkEventConfigure *)event;
	    GDK_PIXMAP_UNREF(d->preview_pm)
	    d->preview_pm = gdk_pixmap_new(
		widget->window, configure->width, configure->height, -1
	    );
	    preview_resize(d);
	    gtk_widget_queue_draw(widget);
	    status = TRUE;
	    break;

	  case GDK_EXPOSE:
	    expose = (GdkEventExpose *)event;
	    preview_draw(d);
	    status = TRUE;
	    break;

	  case GDK_BUTTON_PRESS:
	    button = (GdkEventButton *)event;
	    if(GTK_WIDGET_CAN_FOCUS(widget) && !GTK_WIDGET_HAS_FOCUS(widget))
		gtk_widget_grab_focus(widget);
	    d->preview_button = button->button;
	    d->preview_last_x = (gint)button->x;
	    d->preview_last_y = (gint)button->y;
	    status = TRUE;
	    break;

	  case GDK_BUTTON_RELEASE:
	    button = (GdkEventButton *)event;
	    d->preview_button = 0;
	    status = TRUE;
	    break;

	  case GDK_MOTION_NOTIFY:
	    motion = (GdkEventMotion *)event;
	    if(d->preview_button == 1)
	    {
		const gfloat dtow_coeff = d->preview_dtow_coeff;
		gint x, y, dx, dy;
		GdkModifierType mask;
		GdkWindow *window = motion->window;

		if(motion->is_hint)
		{
		    gdk_window_get_pointer(
			window, &x, &y, &mask
		    );
		}
		else
		{
		    x = (gint)motion->x;
		    y = (gint)motion->y;
		}
		dx = x - d->preview_last_x;
		dy = y - d->preview_last_y;

		d->preview_last_x = x;
		d->preview_last_y = y;

		if(dtow_coeff > 0.0f)
		{
		    if(dx != 0)
		    {
			x = GTK_ENTRY_GET_VALUEI(d->x_spin) + (dx / dtow_coeff);
			gtk_spin_button_set_value(
			    GTK_SPIN_BUTTON(d->x_spin), (gfloat)x
			);
		    }
		    if(dy != 0)
		    {
			y = GTK_ENTRY_GET_VALUEI(d->y_spin) + (dy / dtow_coeff);
			gtk_spin_button_set_value(
			    GTK_SPIN_BUTTON(d->y_spin), (gfloat)y
			);
		    }
		}
	    }
	    status = TRUE;
	    break;
	}

	return(status);
}


/*
 *	Updates the geometry of the Preview GtkDrawingArea and the
 *	preview_dtow_coeff.
 */
static void preview_resize(header_dlg_struct *d)
{
	gint width, height, bwidth, bheight;
	GtkWidget *w = (d != NULL) ? d->preview_da : NULL;
	if(w == NULL)
	    return;

	/* Get size of the preview widget */
	width = w->allocation.width;
	height = w->allocation.height;

	/* Get base size */
	bwidth = GTK_ENTRY_GET_VALUEI(d->base_width_spin);
	bheight = GTK_ENTRY_GET_VALUEI(d->base_height_spin);

	/* If base size is not set then use the image data size */
	if(bwidth <= 0)
	    bwidth = d->width;
	if(bheight <= 0)
	    bheight = d->height;

	/* Reduce base size to accomidate for preview borders */
	bwidth -= 4;
	bheight -= 4;

	/* Recalculate zoom */
	if(bwidth > bheight)
	    d->preview_dtow_coeff = (bwidth > 0) ?
		((gfloat)width / (gfloat)bwidth) : 0.0f;
	else
	    d->preview_dtow_coeff = (bheight > 0) ?
		((gfloat)height / (gfloat)bheight) : 0.0f;
}

/*
 *	Redraws the Preview GtkDrawingArea.
 */
static void preview_draw(header_dlg_struct *d)
{
	gfloat baspect, dtow_coeff;
	gint width, height, x, y, bx, by, bwidth_orig, bwidth, bheight;
	GdkPixmap *pixmap;
	GdkWindow *window;
	GdkDrawable *drawable;
	GtkStateType state;
	GtkStyle *style;
	GtkWidget *w = (d != NULL) ? d->preview_da : NULL;
	if(w == NULL)
	    return;

	width = w->allocation.width;
	height = w->allocation.height;
	pixmap = d->preview_pm;
	window = w->window;
	state = GTK_WIDGET_STATE(w);
	style = gtk_widget_get_style(w);
	if((pixmap == NULL) || (window == NULL) || (style == NULL) ||
	   (width <= 0) || (height <= 0)
	)
	    return;

	drawable = pixmap;

	/* Draw background */
	gdk_draw_rectangle(
	    drawable, style->bg_gc[state], TRUE,
	    0, 0, width, height
	);

	/* Get base geometry */
	bwidth_orig = bwidth = GTK_ENTRY_GET_VALUEI(d->base_width_spin);
	bheight = GTK_ENTRY_GET_VALUEI(d->base_height_spin);
	if(bwidth <= 0)
	    bwidth_orig = bwidth = d->width;
	if(bheight <= 0)
	    bheight = d->height;

	/* Calculate aspect of base */
	baspect = (bheight > 0) ?
	    ((gfloat)bwidth / (gfloat)bheight) : 0.0f;

	/* Calculate base geometry in window coordinates */
	if(bwidth > bheight)
	{
	    bwidth = width;
	    bheight = (baspect > 0.0f) ?
		(bwidth / baspect) : bwidth;
	    bx = 0;
	    by = (height / 2) - (bheight / 2);
	}
	else
	{
	    bheight = height;
	    bwidth = (baspect > 0.0f) ?
		(bheight * baspect) : bheight;
	    bx = (width / 2) - (bwidth / 2);
	    by = 0;
	}
	if((bwidth <= 5) || (bwidth <= 5))
	    return;

	/* Draw base frame */
	gtk_draw_shadow(
	    style, drawable, state, GTK_SHADOW_IN,
	    bx, by, bwidth, bheight
	);

	/* Decrease base size due to drawing frame */
	bx += 2;
	by += 2;
	bwidth -= 4;
	bheight -= 4;

	/* Draw background */
	gdk_draw_rectangle(
	    drawable, style->base_gc[state], TRUE,
	    bx, by, bwidth, bheight
	);                     

	/* Get data to window coefficient */
	dtow_coeff = d->preview_dtow_coeff;

	/* Calculate thumb position in window coordinates */
	x = (GTK_ENTRY_GET_VALUEI(d->x_spin) * dtow_coeff) + bx;
	y = (GTK_ENTRY_GET_VALUEI(d->y_spin) * dtow_coeff) + by;

	/* Draw thumb */
	if((d->thumb_rgba != NULL) && (d->rgba != NULL))
	{
	    gint	src_width = d->width,
			src_height = d->height,
			src_bpl = d->bpl,
			thumb_width = MIN(
		(gint)(src_width * dtow_coeff), d->thumb_width
	    ),
			thumb_height = MIN(
		(gint)(src_height * dtow_coeff), d->thumb_height
	    );
	    guint8 *thumb_rgba = d->thumb_rgba;
	    const guint8 *src_rgba = d->rgba;
	    GdkRectangle rect;
	    GdkGC *gc = style->fg_gc[state];

	    rect.x = bx;
	    rect.y = by;
	    rect.width = bwidth;
	    rect.height = bheight;
	    gdk_gc_set_clip_rectangle(gc, &rect);

	    GUIImageBufferResize(
		4,
		src_rgba, src_width, src_height, src_bpl,
		thumb_rgba, thumb_width, thumb_height, -1
	    );
	    gdk_draw_rgb_32_image(
		drawable, gc,
		x, y,
		thumb_width, thumb_height,
		GDK_RGB_DITHER_NORMAL,
		(guchar *)thumb_rgba,
		thumb_width * 4
	    );

	    gdk_gc_set_clip_rectangle(gc, NULL); 
	}

	/* Send drawable to window if drawable is not the window */
	if(drawable != window)
	    gdk_draw_pixmap(
		window, style->fg_gc[state], drawable,
		0, 0, 0, 0, width, height
	    );
}

/*
 *	Builds a standard toplevel dialog and returns the pointers to
 *	the created widgets.
 */
static void build_std_dialog(
	gint width, gint height, const gchar *title,
	GtkWidget *ref_toplevel,
	GtkWidget **toplevel_rtn, GtkWidget **client_vbox_rtn,
	GtkAccelGroup **accelgrp_rtn
)
{
	gint	bw = GUI_BUTTON_HLABEL_WIDTH_DEF,
		bh = GUI_BUTTON_HLABEL_HEIGHT_DEF,
		border_major = 5;
	GdkWindow *window;
	GtkAccelGroup *accelgrp;
	GtkWidget *w, *parent, *parent2;
	GtkWidget *toplevel, *main_vbox;


	/* Create keyboard accelerator group */
	accelgrp = gtk_accel_group_new();
	if(accelgrp_rtn != NULL)
	    *accelgrp_rtn = accelgrp;

	/* Create toplevel */
	toplevel = parent = w = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_widget_set_usize(w, width, height);
	gtk_window_set_title(GTK_WINDOW(w), title);
#ifdef PROG_NAME
	gtk_window_set_wmclass(
	    GTK_WINDOW(w), "resize_dialog", PROG_NAME
	);
#endif
	gtk_widget_realize(w);
	window = w->window;
	if(window != NULL)
	{
	    gdk_window_set_decorations(
		window,
		GDK_DECOR_BORDER | GDK_DECOR_TITLE | GDK_DECOR_MENU |
		GDK_DECOR_MINIMIZE
	    );
	    gdk_window_set_functions(
		window,
		GDK_FUNC_MOVE | GDK_FUNC_MINIMIZE | GDK_FUNC_CLOSE
	    );
	    GUISetWMIcon(window, (guint8 **)icon_edit_48x48_xpm);
	}
	gtk_signal_connect(
	    GTK_OBJECT(w), "delete_event",
	    GTK_SIGNAL_FUNC(close_cb), toplevel
	);
	gtk_container_border_width(GTK_CONTAINER(w), 0);
	gtk_accel_group_attach(accelgrp, GTK_OBJECT(w));
	if((ref_toplevel != NULL) ?
	    GTK_IS_WINDOW(GTK_OBJECT(ref_toplevel)) : FALSE
	)
	{
	    gtk_window_set_modal(GTK_WINDOW(toplevel), TRUE);
	    gtk_window_set_transient_for(
		GTK_WINDOW(toplevel), GTK_WINDOW(ref_toplevel)
	    );
	}
	if(toplevel_rtn != NULL)
	    *toplevel_rtn = toplevel;

	main_vbox = w = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(parent), w);
	gtk_widget_show(w);
	parent = w;

	w = gtk_vbox_new(FALSE, border_major);
	gtk_container_border_width(GTK_CONTAINER(w), border_major);
	gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	gtk_widget_show(w);
	parent2 = w;
	if(client_vbox_rtn != NULL)
	    *client_vbox_rtn = w;

	w = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(main_vbox), w, FALSE, FALSE, 0);
	gtk_widget_show(w);

	/* Buttons */
	w = gtk_hbox_new(TRUE, 0);
	gtk_box_pack_start(GTK_BOX(main_vbox), w, FALSE, FALSE, border_major);
	gtk_widget_show(w);
	parent2 = w;

	/* Set Button */
	w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_ok_20x20_xpm,
	    "Set",
	    NULL
	);
	gtk_widget_set_usize(w, bw, bh);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT | GTK_HAS_DEFAULT);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(set_btn_cb), toplevel
	);
	gtk_accel_group_add(
	    accelgrp, GDK_Return, 0, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	gtk_accel_group_add(
	    accelgrp, GDK_3270_Enter, 0, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	gtk_accel_group_add(
	    accelgrp, GDK_KP_Enter, 0, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	gtk_accel_group_add(
	    accelgrp, GDK_ISO_Enter, 0, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	gtk_widget_show(w);

	/* Cancel Button */
	w = GUIButtonPixmapLabelH(
	    (guint8 **)icon_cancel_20x20_xpm,
	    "Cancel",
	    NULL
	);
	gtk_widget_set_usize(w, bw, bh);
	GTK_WIDGET_SET_FLAGS(w, GTK_CAN_DEFAULT);
	gtk_box_pack_start(GTK_BOX(parent2), w, TRUE, FALSE, 0);
	gtk_signal_connect(
	    GTK_OBJECT(w), "clicked",
	    GTK_SIGNAL_FUNC(cancel_btn_cb), toplevel
	);
	gtk_accel_group_add(
	    accelgrp, GDK_Escape, 0, GTK_ACCEL_VISIBLE,
	    GTK_OBJECT(w), "clicked"
	);
	gtk_widget_show(w);
}


/*
 *	Queries user to edit header.
 *
 *	Returns TRUE on Set or FALSE on Cancel.
 */
gboolean iv_edit_header(
	GtkWidget *ref_toplevel,
	const gchar *filename,
	const guint8 *rgba,
	gint width, gint height, gint bpl,
	gchar **title, gchar **author, gchar **comments,
	gint *x, gint *y,
	gint *base_width, gint *base_height
)
{
	gint	border_major = 5,
		border_minor = 2;
	gchar *s;
	GtkAccelGroup *accelgrp = NULL;
	GtkWidget *toplevel = NULL, *client_vbox = NULL;
	header_dlg_struct *d;

	/* Reset globals */
	got_response = FALSE;

	/* Create dialog base widgets */
	s = g_strdup_printf(
	    "Edit Header: %s",
	    STRISEMPTY(filename) ? "Untitled" : filename
	);
	build_std_dialog(
	    400, -1,
	    s,
	    ref_toplevel,
	    &toplevel, &client_vbox, &accelgrp
	);
	g_free(s);
	d = HEADER_DLG(g_malloc0(sizeof(header_dlg_struct)));
	d->toplevel = toplevel;
	d->accelgrp = accelgrp;
	d->freeze_count = 0;

	d->preview_button = 0;
	d->preview_last_x = 0;
	d->preview_last_y = 0;
	d->preview_dtow_coeff = 0.0f;

	d->rgba = rgba;
	d->width = width;
	d->height = height;
	d->bpl = bpl;


	/* Create Header Dialog widgets */
	if(client_vbox != NULL)
	{
	    gint	label_width, value_width,
			preview_width = 100,
			preview_height = 100;
	    GdkWindow *window = toplevel->window;
	    GtkAdjustment *adj;
	    GtkStyle *style = gtk_widget_get_style(toplevel);
	    GtkWidget	*w,
			*parent = client_vbox,
			*parent2, *parent3, *parent4;


	    w = gtk_hbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent = w;


	    /* Left column vbox */
	    w = gtk_vbox_new(TRUE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent2 = w;
	    /* Icon */
	    w = gtk_pixmap_new_from_xpm_d(
		window,
		(style != NULL) ? &style->bg[GTK_STATE_NORMAL] : NULL,
		(guint8 **)icon_edit_32x32_xpm
	    );
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);   


	    /* Right column vbox */
	    w = gtk_vbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent), w, TRUE, TRUE, 0);
	    gtk_widget_show(w);
	    parent2 = w;

	    label_width = 60;

	    /* Hbox for Title */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;
	    /* Label */
	    w = gtk_alignment_new(1.0f, 0.5f, 0.0f, 0.0f);
	    gtk_widget_set_usize(w, label_width, -1);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent4 = w;
	    w = gtk_label_new("Title:");
	    gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);
	    gtk_container_add(GTK_CONTAINER(parent4), w);
	    gtk_widget_show(w);
	    /* Entry */
	    d->title_entry = w = gtk_entry_new();
	    gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	    GUISetWidgetTip(w,
"Enter the title of this image"
	    );
	    gtk_widget_show(w);

	    /* Hbox for Author */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;
	    /* Label */
	    w = gtk_alignment_new(1.0f, 0.5f, 0.0f, 0.0f);
	    gtk_widget_set_usize(w, label_width, -1);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent4 = w;
	    w = gtk_label_new("Author:");
	    gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);
	    gtk_container_add(GTK_CONTAINER(parent4), w);
	    gtk_widget_show(w);
	    /* Entry */
	    d->author_entry = w = gtk_entry_new();
	    gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	    GUISetWidgetTip(w,
"Enter the name of the author of this image"
	    );
	    gtk_widget_show(w);

	    /* Hbox for Comments */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;      
	    /* Label */ 
	    w = gtk_alignment_new(1.0f, 0.5f, 0.0f, 0.0f);
	    gtk_widget_set_usize(w, label_width, -1);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent4 = w;      
	    w = gtk_label_new("Comments:");
	    gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_RIGHT);
	    gtk_container_add(GTK_CONTAINER(parent4), w);
	    gtk_widget_show(w);      
	    /* Entry */
	    d->comments_entry = w = gtk_entry_new();
	    gtk_box_pack_start(GTK_BOX(parent3), w, TRUE, TRUE, 0);
	    GUISetWidgetTip(w,
"Enter the comments or description of this image"
	    );
	    gtk_widget_show(w);


	    w = gtk_hseparator_new();
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);


	    /* Hbox for Position & Base Size */
	    w = gtk_hbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;

	    value_width = 60;

	    /* Hbox for X Position */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent4 = w;
	    /* Label */
	    w = gtk_label_new("X:");
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    /* Spin Button */
	    adj = (GtkAdjustment *)gtk_adjustment_new(
		0.0f,
		(gfloat)-((guint16)-1),
		(gfloat)((guint16)-1),
		1.0f,
		10.0f,
		0.0f  
	    );
	    gtk_signal_connect(
		GTK_OBJECT(adj), "value_changed",
		GTK_SIGNAL_FUNC(adj_value_changed_cb), d
	    );
	    d->x_spin = w = gtk_spin_button_new(adj, 1.0, 0);
	    gtk_widget_set_usize(w, value_width, -1);
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    GUISetWidgetTip(w,
"Set the x coordinate offset of this image on its base (in pixels)"
	    );
	    gtk_widget_show(w);

	    /* Hbox for Y Position */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent4 = w;
	    /* Label */
	    w = gtk_label_new("Y:");
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_widget_show(w); 
	    /* Spin Button */
	    adj = (GtkAdjustment *)gtk_adjustment_new(
		0.0f,
		(gfloat)-((guint16)-1),
		(gfloat)((guint16)-1),
		1.0f,
		10.0f,
		0.0f  
	    );
	    gtk_signal_connect(
		GTK_OBJECT(adj), "value_changed",
		GTK_SIGNAL_FUNC(adj_value_changed_cb), d
	    );
	    d->y_spin = w = gtk_spin_button_new(adj, 1.0, 0);
	    gtk_widget_set_usize(w, value_width, -1);
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    GUISetWidgetTip(w,
"Set the y coordinate offset of this image on its base (in pixels)"
	    );
	    gtk_widget_show(w);

	    /* Hbox for Base Width */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent4 = w;
	    /* Label */
	    w = gtk_label_new("BW:");
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    /* Spin Button */
	    adj = (GtkAdjustment *)gtk_adjustment_new(
		0.0f,
		0.0f,
		(gfloat)((guint16)-1),
		1.0f,
		10.0f,
		0.0f
	    );
	    gtk_signal_connect(
		GTK_OBJECT(adj), "value_changed",
		GTK_SIGNAL_FUNC(adj_value_changed_cb), d
	    );
	    d->base_width_spin = w = gtk_spin_button_new(adj, 1.0, 0);
	    gtk_widget_set_usize(w, value_width, -1);
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    GUISetWidgetTip(w,
"Set the width of the base (in pixels)"
	    );
	    gtk_widget_show(w);

	    /* Hbox for Base Height */
	    w = gtk_hbox_new(FALSE, border_minor);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent4 = w;
	    /* Label */
	    w = gtk_label_new("BH:");
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    /* Spin Button */
	    adj = (GtkAdjustment *)gtk_adjustment_new(
		0.0f,
		0.0f,
		(gfloat)((guint16)-1),
		1.0f,
		10.0f,
		0.0f
	    );
	    gtk_signal_connect(
		GTK_OBJECT(adj), "value_changed",
		GTK_SIGNAL_FUNC(adj_value_changed_cb), d
	    );
	    d->base_height_spin = w = gtk_spin_button_new(adj, 1.0, 0);
	    gtk_widget_set_usize(w, value_width, -1);
	    gtk_box_pack_start(GTK_BOX(parent4), w, FALSE, FALSE, 0);
	    GUISetWidgetTip(w,
"Set the height of the base (in pixels)"
	    );
	    gtk_widget_show(w);

	    /* Layout Frame */
	    w = gtk_frame_new("Layout");
	    gtk_frame_set_shadow_type(GTK_FRAME(w), GTK_SHADOW_ETCHED_IN);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;
	    /* Hbox for Preview */
	    w = gtk_hbox_new(TRUE, 0);
	    gtk_container_border_width(GTK_CONTAINER(w), border_major);
	    gtk_container_add(GTK_CONTAINER(parent3), w);
	    gtk_widget_show(w);
	    parent3 = w;

	    /* Preview Drawing Area */
	    d->preview_da = w = gtk_drawing_area_new();
	    gtk_widget_set_usize(w, preview_width, preview_height);
	    gtk_widget_add_events(
		w,
		GDK_STRUCTURE_MASK | GDK_EXPOSURE_MASK |
		GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
		GDK_POINTER_MOTION_MASK | GDK_POINTER_MOTION_HINT_MASK
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "configure_event",
		GTK_SIGNAL_FUNC(preview_event_cb), d 
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "expose_event",
		GTK_SIGNAL_FUNC(preview_event_cb), d
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "button_press_event",
		GTK_SIGNAL_FUNC(preview_event_cb), d
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "button_release_event",
		GTK_SIGNAL_FUNC(preview_event_cb), d
	    );
	    gtk_signal_connect(
		GTK_OBJECT(w), "motion_notify_event",
		GTK_SIGNAL_FUNC(preview_event_cb), d
	    );
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_realize(w);
	    gtk_widget_show(w);

	    /* Preview thumb data
	     *
	     * Allocate it to twice the size of the preview
	     * GtkDrawingArea for reasonable (but not perfect) x2 zoom
	     * coverage
	     */
	    d->thumb_width = preview_width * 2;
	    d->thumb_height = preview_height * 2;
	    d->thumb_rgba = (guint8 *)g_malloc(
		d->thumb_width * d->thumb_height * 4
	    );


	    /* Limitations notice */
	    w = gtk_hbox_new(FALSE, border_major);
	    gtk_box_pack_start(GTK_BOX(parent2), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    parent3 = w;
	    w = gtk_pixmap_new_from_xpm_d(
		window,
		(style != NULL) ? &style->bg[GTK_STATE_NORMAL] : NULL,
		(guint8 **)icon_info_20x20_xpm
	    );
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);
	    w = gtk_label_new(
"Not all values are supported by all image formats, verify\n\
that the image format supports the value(s) in question\n\
before editing and saving."
	    );
	    gtk_label_set_justify(GTK_LABEL(w), GTK_JUSTIFY_LEFT);
	    gtk_box_pack_start(GTK_BOX(parent3), w, FALSE, FALSE, 0);
	    gtk_widget_show(w);



	    /* Set initial values */
	    if((title != NULL) ? !STRISEMPTY(*title) : FALSE)
		gtk_entry_set_text(
		    GTK_ENTRY(d->title_entry), *title
		);
	    if((author != NULL) ? !STRISEMPTY(*author) : FALSE)
		gtk_entry_set_text(
		    GTK_ENTRY(d->author_entry), *author
		);
	    if((comments != NULL) ? !STRISEMPTY(*comments) : FALSE)
		gtk_entry_set_text(
		    GTK_ENTRY(d->comments_entry), *comments
		);
	    gtk_spin_button_set_value(
		GTK_SPIN_BUTTON(d->x_spin),
		(x != NULL) ? (gfloat)*x : 0.0f
	    );
	    gtk_spin_button_set_value(
		GTK_SPIN_BUTTON(d->y_spin),
		(y != NULL) ? (gfloat)*y : 0.0f
	    );
	    gtk_spin_button_set_value(
		GTK_SPIN_BUTTON(d->base_width_spin),
		(base_width != NULL) ? (gfloat)*base_width : width
	    );
	    gtk_spin_button_set_value(
		GTK_SPIN_BUTTON(d->base_height_spin),
		(base_height != NULL) ? (gfloat)*base_height : height
	    );
	}


	/* Map toplevel */
	gtk_widget_show_raise(toplevel);


	/* Push main loop to wait for user response */
	gtk_main();


	/* Get values from widgets if user clicked on save */
	if(got_response)
	{
	    if(title != NULL)
	    {
		g_free(*title);
		*title = STRDUP(gtk_entry_get_text(
		    GTK_ENTRY(d->title_entry)
		));
	    }
	    if(author != NULL)
	    {
		g_free(*author);
		*author = STRDUP(gtk_entry_get_text(
		    GTK_ENTRY(d->author_entry)
		));
	    }
	    if(comments != NULL)
	    {
		g_free(*comments);
		*comments = STRDUP(gtk_entry_get_text(
		    GTK_ENTRY(d->comments_entry)
		));
	    }
	    if(x != NULL)
		*x = GTK_ENTRY_GET_VALUEI(d->x_spin);
	    if(y != NULL)
		*y = GTK_ENTRY_GET_VALUEI(d->y_spin);
	    if(base_width != NULL)
		*base_width = GTK_ENTRY_GET_VALUEI(d->base_width_spin);
	    if(base_height != NULL)
		*base_height = GTK_ENTRY_GET_VALUEI(d->base_height_spin);
	}


	/* Destroy widgets and related resources */
	gtk_window_set_modal(GTK_WINDOW(toplevel), FALSE);
	gtk_window_set_transient_for(GTK_WINDOW(toplevel), NULL);

	GTK_WIDGET_DESTROY(d->toplevel)
	GDK_PIXMAP_UNREF(d->preview_pm)
	GTK_ACCEL_GROUP_UNREF(d->accelgrp)
	g_free(d->thumb_rgba);
	g_free(d);

	/* Return user response */
	return(got_response);
}
